///////////////////////////////////////////////
//                                           //
// VRLab Mental Vision Module 1: Hermite     //
// Author: Achille Peternier                 //
// Copyright: EPFL-VRLab 2005 (C)            //
//                                           //
///////////////////////////////////////////////
#include "../data/common.h"
#pragma comment(lib, "opengl32.lib")



//////////////
// #DEFINES //
//////////////
#define MODULE_NAME "Hermite interpolation"


//////////////
// ACCURACY //
//////////////
#define ACCURACY 40  // Lines per fragment
#define POINTS   8   // Control points



////////////////////////////////////////////////////////////////////////////////////////////////////
MVVECTOR3 splinePointHermite(float s, MVVECTOR3 &P1, MVVECTOR3 &P2, MVVECTOR3 &T1, MVVECTOR3 &T2) //
{ 
   float h1 =  2*pow(s, 3) - 3*pow(s, 2) + 1;          // calculate basis function 1
   float h2 = -2*pow(s, 3) + 3*pow(s, 2);              // calculate basis function 2
   float h3 = pow(s, 3) - 2*pow(s, 2) + s;           // calculate basis function 3
   float h4 = pow(s, 3) -   pow(s, 2);                // calculate basis function 4
   return P1*h1 +                                      // multiply and sum all funtions
          P2*h2 +                                      // together to build the interpolated
          T1*h3 +                                      // point along the curve.
          T2*h4;
}



//////////
// MAIN //
//////////

int main(int argc, char *argv[])
{   
   /////////////////
   // Init MVisio //
   /////////////////
   INIT_MVISIO(MODULE_NAME, argc, argv);
   
   // Common interface:
   USE_FONT();  
   USE_BACKGROUND(MODULE_SIZEX, MODULE_SIZEY);
   USE_QUITBUTTON(MODULE_SIZEX, MODULE_SIZEY);
  
   
   /////////////
   // Hermite //
   /////////////     

   // Interface:
   MVBUTTON *point[POINTS];
   for (int c=0; c<POINTS; c++)
   {
         point[c] = new MVBUTTON();
         point[c]->setFont(font);
         point[c]->setSizes(16, 16);
         point[c]->setStaticStatus(false);
         point[c]->setLimitStatus(true);
         point[c]->setLimits(5, 44, MODULE_SIZEX-5, MODULE_SIZEY-95);
         char _pname[20];
         sprintf(_pname, "%d", c+1);
         point[c]->setCaption(_pname);
         point[c]->setColor(0.0f, 0.8f, 0.0f, 0.5f);
         point[c]->setPosition((float) ((c+1)*MODULE_SIZEX/(POINTS+1)), (float) (MODULE_SIZEY/2));
   }
   
   // Control points:
   MVVECTOR3 cPoint[POINTS]; 
   EXF_ZEROMEMORY(cPoint, sizeof(MVVECTOR3)*POINTS);
   
   
   ////////////////////////////
   // Point modifier interface:
   MVWINDOW *modifyPointWin = new MVWINDOW();
   modifyPointWin->setFont(font);
   modifyPointWin->setSizes(200, 58);  
   modifyPointWin->setCaption("Point modifier");
   modifyPointWin->setTexture(baseTexture);
   modifyPointWin->setPosition((float) (MODULE_SIZEX-MODULE_SIZEX/3), (float) (MODULE_SIZEY-MODULE_SIZEY/3));
   
   // Tangent:
   MVHSCROLLBAR *modifyPointScroll1 = new MVHSCROLLBAR();   
   modifyPointScroll1->setPosition(120, 10);
   modifyPointScroll1->link(modifyPointWin);
   modifyPointScroll1->setMinValue(-180.0f);
   modifyPointScroll1->setMaxValue(180.0f);
   modifyPointScroll1->setValue(0.0f);
   
   // Intensity:
   MVHSCROLLBAR *modifyPointScroll2 = new MVHSCROLLBAR();   
   modifyPointScroll2->setPosition(120, 31);
   modifyPointScroll2->link(modifyPointWin);
   modifyPointScroll2->setMinValue(0.0f);   
   modifyPointScroll2->setMaxValue(1000.0f);   
   modifyPointScroll2->setValue(0.0f);
   int currentSelected = 0;
  
   // Module slide:
   MVTEXTURE *slideTexture = new MVTEXTURE();
   slideTexture->load("hermiteinterpolation_slide.tga");
   MVWINDOW *slideWindow = new MVWINDOW();
   slideWindow->setFont(font);
   slideWindow->setCaption("Slide");
   slideWindow->setSizes(512, 440);
   slideWindow->setColor(1.0f, 1.0f, 1.0f, 1.0f);
   slideWindow->setTexture(slideTexture);
   slideWindow->setPosition(15, 85);
   
   MVVSCROLLBAR *slideScroll = new MVVSCROLLBAR();
   slideScroll->setMinValue(0.1f);
   slideScroll->setMaxValue(1.0f);
   slideScroll->setValue(1.0f);
   slideScroll->setPosition(495, 12);
   slideScroll->link(slideWindow);            

///////////////
// Mainloop: //
///////////////   
   
   int mouseX = 0, mouseY = 0;
   bool mouseBL = false, mouseBR = false;
   bool done = false;
   while (!done)   
	{		      
         // Update values:
	      char caption[50];
	      sprintf(caption, "Current point: %d", currentSelected+1);
	      modifyPointWin->setCaption(caption);
	      char values[255];	      
	      sprintf(values, "^y0005^x0010tangent: ^x0003%.1f\n^y0005^x0010intensity ^x0003%.1f\n", 
	              modifyPointScroll1->getValue(), 
	              modifyPointScroll2->getValue());	           
	      modifyPointWin->clear();	      
	      modifyPointWin->write(values);	      
         modifyPointWin->format();
	
	      // Rendering:	   
	   	MVISIO::clear(true, true, false);	   		   	
			MVISIO::begin2D();			   
			   PASS_BACKGROUND;					     			  
			MVISIO::end2D();			
			
			// Draw spline:			
			MVISIO::begin2D();		
			   // Title:
			   font->putString(10, 13, MODULE_NAME);	  
			   glDisable(GL_TEXTURE_2D);
			   
			   // Lines:
			   glLineWidth(LINE_WIDTH);	
			   glBegin(GL_LINE_STRIP);			   
			   for (int c=0; c<POINTS-1; c++)
			         for (int d=0; d<=ACCURACY; d++)
			         {
			               // Time interval:
			               float t = (float)d / (float)ACCURACY;
			               
			               // Start-end point (centered on buttons):
			               MVVECTOR3 p1 = point[c]->getPosition();
			               MVVECTOR3 p2 = point[c+1]->getPosition();
			               p1 += MVVECTOR3((float) point[c]->getSizeX()/2.0f, (float) point[c]->getSizeY()/2.0f, 0.0f);
			               p2 += MVVECTOR3((float) point[c+1]->getSizeX()/2.0f, (float) point[c+1]->getSizeY()/2.0f, 0.0f);
			               
			               // Tangents:
			               float alpha = E3M_TORAD(cPoint[c].x);
			               float intensity = cPoint[c].y;			               
			               MVVECTOR3 t1 = MVVECTOR3((float) (intensity*cos(alpha)), (float) (intensity*sin(alpha)), 0.0f);
			               alpha = E3M_TORAD(cPoint[c+1].x);
			               intensity = cPoint[c+1].y;			               
			               MVVECTOR3 t2 = MVVECTOR3((float) (intensity*cos(alpha)), (float) (intensity*sin(alpha)), 0.0f);	
			               
			               // Compute and display curbe:		               
			               MVVECTOR3 p = splinePointHermite(t, p1, p2, t1, t2);
			               glColor4f(1.0f, 1.0f, 0.0f, 1.0f);			      
			               glVertex2f(p.x, p.y);			           
			         }
			   glEnd();		
			   
			   // Tangents:
			   glBegin(GL_LINES);  
			   for (int c=0; c<POINTS; c++)
	         {
			               // Time interval:
			               float t = 0.0f;
			               
			               // Start-end point (centered on buttons):
			               MVVECTOR3 p1 = point[c]->getPosition();			               
			               p1 += MVVECTOR3((float) point[c]->getSizeX()/2.0f, (float) point[c]->getSizeY()/2.0f, 0.0f);
			               
			               // Tangent:
			               float alpha = E3M_TORAD(cPoint[c].x);
			               float intensity = cPoint[c].y/3.0f;			               
			               MVVECTOR3 t1 = MVVECTOR3((float) (intensity*cos(alpha)), (float) (intensity*sin(alpha)), 0.0f);			               
			               
			               // Compute and display curbe:		               
			               glColor4f(1.0f, 0.3f, 1.0f, 0.9f);			      
			               glVertex2f(p1.x, p1.y);			               
			               glColor4f(0.5f, 0.25f, 0.5f, 0.6f);			      
			               glVertex2f(p1.x+t1.x, p1.y+t1.y);			               			               			               
			         }
			   glEnd();				   
			MVISIO::end2D();
			
			// Draw control points:
			MVISIO::begin2D();
			   PASS_QUITBUTTON;
			   modifyPointWin->pass();			
			   for (int c=0; c<POINTS; c++)
			         point[c]->pass();		
			   slideWindow->setColor(1.0f, 1.0f, 1.0f, slideScroll->getValue());
			   slideWindow->pass();	  
			MVISIO::end2D();
			
			////////////////
			// Check events:			
			SDL_Event event;			
			//SDL_PumpEvents();						
			while (SDL_PollEvent(&event)) 
			{
					// Quit by click on X:
					if (event.type == SDL_QUIT) 
							done = true;
               
               // Resize?
				   if (event.type == SDL_VIDEORESIZE)
				         MVSETTINGS::resize(0, 0, event.resize.w, event.resize.h);														
               
               // Quit by pressing ESC:
					if (event.type == SDL_KEYDOWN) 
					{
							// Quit by ESC:
							if (event.key.keysym.sym == SDLK_ESCAPE) 
							       done = true;															
			      }																
			}         
			
			// Update mouse:
			if (SDL_GetMouseState(&mouseX, &mouseY)&SDL_BUTTON(1))
					mouseBL = true;
			else 
					mouseBL = false;
			
			// Check MVisio events:
			MVID id=0;
			MVELEMENT::manageGUIEvents(0, mouseX, mouseY, mouseBL, mouseBR, &id);  
			if (PRESSED_QUITBUTTON(id))
			      done = true;
			for (int c=0; c<POINTS; c++)
			      if (id == point[c]->getId())
			      {
			            if (currentSelected != c)
			            {
			                  currentSelected = c;
			                  modifyPointScroll1->setValue(cPoint[currentSelected].x);
			                  modifyPointScroll2->setValue(cPoint[currentSelected].y);			
			            }
			      }
			            
         // Apply sliders:
         cPoint[currentSelected].x = modifyPointScroll1->getValue();
         cPoint[currentSelected].y = modifyPointScroll2->getValue();         
			
			// Display:		
			MVISIO::swap();		
			Sleep(20); // Don't kill mthreading ;)
	}      
   
   // Done:
   MVISIO::free();
   return 0;
}
